--[[ FlyoutListButton ]]

local StoredCursor = {}
local ParentAlpha
local QUICK_SLOT = "Interface\\Buttons\\UI-Quickslot"
local QUICK_SLOT_2 = "Interface\\Buttons\\UI-Quickslot2"
local QUESTION_MARK = "Interface\\Icons\\INV_Misc_QuestionMark"
FlyoutListButton = {}


local OnClickWrapScript_HideSet = [=[
	local listFrame = self:GetParent()
	if listFrame then
		local arrowBtn = listFrame:GetParent()
		if arrowBtn then
			arrowBtn:SetAttribute("expanded", false)
			listFrame:Hide()
			
			local actionButton = arrowBtn:GetParent()
			local attr_type = self:GetAttribute("type")
			if (attr_type == "spell") or (attr_type == "item") or (attr_type == "macro") then
				actionButton:SetAttribute("type", attr_type)
				actionButton:SetAttribute(attr_type, self:GetAttribute(attr_type))
				self:SetAttribute("override_pending", true)
				actionButton:SetAttribute("fbc_tooltip_source", self:GetName())
			end
		end
	end	
]=]

local OnClickWrapScript_Hide = [=[
	local listFrame = self:GetParent()
	if listFrame then
		local arrowBtn = listFrame:GetParent()
		if arrowBtn then
			arrowBtn:SetAttribute("expanded", false)
			listFrame:Hide()
		end
	end	
]=]

local OnClickWrapScript_Set = [=[
	local listFrame = self:GetParent()
	if listFrame then
		local arrowBtn = listFrame:GetParent()
		if arrowBtn then
			local actionButton = arrowBtn:GetParent()
			local attr_type = self:GetAttribute("type")
			if (attr_type == "spell") or (attr_type == "item") or (attr_type == "macro") then
				actionButton:SetAttribute("type", attr_type)
				actionButton:SetAttribute(attr_type, self:GetAttribute(attr_type))
				self:SetAttribute("override_pending", true)
				actionButton:SetAttribute("fbc_tooltip_source", self:GetName())
			end
		end
	end	
]=]

-- presets

local preset_by_companion = {
	["UpdateText"] 		= "EmptyFunc",
	["UpdateCount"] 	= "EmptyFunc",
	["SetTooltip"] 		= "SetTooltipCompanion",
	["UpdateTooltip"] 	= "UpdateTooltipCompanion",
	["GetTexture"] 		= "GetTextureCompanion",
	["UpdateChecked"] 	= "UpdateCheckedCompanion",
	["UpdateEquipped"] 	= "UpdateEquippedNotItem",
	["UpdateCooldown"] 	= "UpdateCooldownNone",
	["GetUsable"] 		= "GetUsableComapnion",
	["InRange"] 		= "InRangeCompanion",
	["GlowShow"]		= "EmptyFunc",
	["GlowHide"]		= "EmptyFunc",
}

local preset_by_spell = {
	["UpdateText"] 		= "EmptyFunc",
	["UpdateCount"] 	= "UpdateCountSpell",
	["SetTooltip"] 		= "SetTooltipSpell",
	["UpdateTooltip"] 	= "UpdateTooltipSpell",
	["GetTexture"] 		= "GetTextureSpell",
	["UpdateChecked"] 	= "UpdateCheckedSpell",
	["UpdateEquipped"] 	= "UpdateEquippedNotItem",
	["UpdateCooldown"] 	= "UpdateCooldownSpell",
	["GetUsable"] 		= "GetUsableSpell",
	["InRange"] 		= "InRangeSpell",
	["GlowShow"]		= "SpellGlowShow",
	["GlowHide"]		= "SpellGlowHide",
}

local preset_by_item = {
	["UpdateText"] 		= "EmptyFunc",
	["UpdateCount"] 	= "UpdateCountItem",
	["SetTooltip"] 		= "SetTooltipItem",
	["UpdateTooltip"] 	= "UpdateTooltipItem",
	["GetTexture"] 		= "GetTextureItem",
	["UpdateChecked"] 	= "UpdateCheckedItem",
	["UpdateEquipped"] 	= "UpdateEquippedItem",
	["UpdateCooldown"] 	= "UpdateCooldownItem",
	["GetUsable"] 		= "GetUsableItem",
	["InRange"] 		= "InRangeItem",
	["GlowShow"]		= "EmptyFunc",
	["GlowHide"]		= "EmptyFunc",
}

local preset_by_macro = {
	["UpdateText"] 		= "UpdateTextMacro",
	["UpdateCount"] 	= "EmptyFunc",
	["SetTooltip"] 		= "SetTooltipMacro",
	["UpdateTooltip"]	= "UpdateTooltipMacro",
	["GetTexture"] 		= "GetTextureMacro",
	["UpdateChecked"] 	= "UpdateCheckedMacro",
	["UpdateEquipped"] 	= "UpdateEquippedNotItem",
	["UpdateCooldown"] 	= "UpdateCooldownMacro",
	["GetUsable"] 		= "GetUsableMacro",
	["InRange"] 		= "InRangeMacro",
	["GlowShow"]		= "MacroGlowShow",
	["GlowHide"]		= "MacroGlowHide",
}

local preset_by_equipmentset = {
	["UpdateText"] 		= "UpdateTextEquipmentset",
	["UpdateCount"] 	= "EmptyFunc",
	["SetTooltip"] 		= "SetTooltipEquipmentset",
	["UpdateTooltip"] 	= "UpdateTooltipEquipmentset",
	["GetTexture"] 		= "GetTextureEquipmentset",
	["UpdateChecked"] 	= "UpdateCheckedEquipmentset",
	["UpdateEquipped"] 	= "UpdateEquippedNotItem",
	["UpdateCooldown"] 	= "UpdateCooldownNone",
	["GetUsable"] 		= "GetUsableEquipmentset",
	["InRange"] 		= "InRangeEquipmentset",
	["GlowShow"]		= "EmptyFunc",
	["GlowHide"]		= "EmptyFunc",
}

local preset_by_battlepet = {
	["UpdateText"] 		= "EmptyFunc",
	["UpdateCount"] 	= "EmptyFunc",
	["SetTooltip"] 		= "SetTooltipBattlepet",
	["UpdateTooltip"] 	= "UpdateTooltipBattlepet",
	["GetTexture"] 		= "GetTextureBattlepet",
	["UpdateChecked"] 	= "UpdateCheckedBattlepet",
	["UpdateEquipped"] 	= "UpdateEquippedNotItem",
	["UpdateCooldown"] 	= "UpdateCooldownNone",
	["GetUsable"] 		= "GetUsableBattlepet",
	["InRange"] 		= "InRangeBattlepet",
	["GlowShow"]		= "EmptyFunc",
	["GlowHide"]		= "EmptyFunc",
}			
			
function FlyoutListButton_New(parent, idx)
	parent.ButtonList[idx] = CreateFrame("CheckButton", parent:GetName().."ListButton"..idx, parent, "CustomFlyoutListButtonTemplate")
	local btn = parent.ButtonList[idx]
	btn.index = idx
	btn.border = _G[btn:GetName().."Border"]
	btn.count = _G[btn:GetName().."Count"]
	btn.name = _G[btn:GetName().."Name"]
	btn.hotkey = _G[btn:GetName().."HotKey"]
	btn.ntexture = _G[btn:GetName().."NormalTexture"]
	btn.flash = _G[btn:GetName().."Flash"]
	btn.flf = parent
	btn.arrBtn = parent:GetParent()
	btn.actnBtn = btn.arrBtn:GetParent()
	btn.actnBtnName = btn.actnBtn:GetName()
	
	-- adding methods to button
	for k, v in pairs(FlyoutListButton) do
		if type(v) == "function" then
			btn[k] = v
		end
	end
	
	btn:SetMovable(true)
	btn:RegisterForDrag("LeftButton")
	btn:SetNormalTexture(QUICK_SLOT)
	btn:SetAttribute("checkselfcast", true)
	btn:SetAttribute("checkfocuscast", true)
	if FbcRightClickSelfCast then
		btn:SetAttribute("unit2", "player") --right-click self cast
	end
	btn:RegisterForClicks("LeftButtonUp", "RightButtonUp")
	btn:SetClickHandler(FbcHideOnClick, FbcLeftClickSet)

	return btn
end

function FlyoutListButton_AttachToList(parent, idx, expandDir)
	local btn = parent.ButtonList[idx]
	if not(btn) then
		btn = FlyoutListButton_New(parent, idx)
		btn:ClearHandlers()	
		if (FbcLBFMasterGroup) then
			FbcLBFMasterGroup:AddButton(btn)
		end
	end
	btn:SetAnchor(parent, idx, expandDir)

	return btn
end

function FlyoutListButton:ClearHandlers()
	for k, v in pairs(preset_by_spell) do
		self[k] = self.EmptyFunc
	end
end

function FlyoutListButton:EmptyFunc()
	return nil
end

function FlyoutListButton:SetClickHandler(hide, click)
	self:UnwrapScript(self, "OnClick")
	if hide then
		if click then
			self:WrapScript(self, "OnClick", OnClickWrapScript_HideSet)
		else
			self:WrapScript(self, "OnClick", OnClickWrapScript_Hide)
		end
	else
		if click then
			self:WrapScript(self, "OnClick", OnClickWrapScript_Set)
		end
	end
end

function FlyoutListButton:SetRightClickSelfCast(value)
	if value then
		self:SetAttribute("unit2", "player") --right-click self cast enabled
	else
		self:SetAttribute("unit2", "") --right-click self cast disabled
	end
end

function FlyoutListButton:SetAnchor(parent, idx, expandDir)
	self:ClearAllPoints()
	if expandDir == FBC_DIR_UP then
		self:SetPoint("BOTTOM", parent, "BOTTOM", 0, idx * FBC_BUTTON_PLACE_OFFSET + (idx - 1) * FBC_BUTTON_PLACE_SIZE)
	elseif expandDir == FBC_DIR_LEFT then
		self:SetPoint("RIGHT", parent, "RIGHT", - (idx * FBC_BUTTON_PLACE_OFFSET + (idx - 1) * FBC_BUTTON_PLACE_SIZE), 0)
	elseif expandDir == FBC_DIR_DOWN then
		self:SetPoint("TOP", parent, "TOP", 0, - (idx * FBC_BUTTON_PLACE_OFFSET + (idx - 1) * FBC_BUTTON_PLACE_SIZE))
	else
		self:SetPoint("LEFT", parent, "LEFT", idx * FBC_BUTTON_PLACE_OFFSET + (idx - 1) * FBC_BUTTON_PLACE_SIZE, 0)
	end
end

-- keybind stuff

function FlyoutListButton:GetHotkey()
	local key = GetBindingKey("CLICK "..self:GetName()..":LeftButton")
	local displayKey = FbcLibKeyBound:ToShortKey(key)
	return displayKey
end

function FlyoutListButton:SetKey(key) -- binds the given key to the given button
	local binding = "CLICK "..self:GetName()..":LeftButton"
	if (key ~= "" and key ~= nil) then
		SetBinding(key, binding)
		self.hotkey:SetText(GetBindingText(key, "KEY_", 1))
		if not self.hotkey.__LBF_SetPoint then
			self.hotkey:ClearAllPoints()
			self.hotkey:SetPoint("TOPLEFT", self, "TOPLEFT", -2, -2)
		end
		self.hotkey:SetVertexColor(0.6, 0.6, 0.6)
		self.hotkey:Show()
	else
		self.hotkey:SetText(RANGE_INDICATOR)
		if not self.hotkey.__LBF_SetPoint then
			self.hotkey:ClearAllPoints()
			self.hotkey:SetPoint("TOPLEFT", self, "TOPLEFT", 1, -2)
		end
		self.hotkey:Hide()
	end
	self.keybind = key

	FlyoutButtonES:Set({self.actnBtnName, GetActiveSpecGroup(), self.index, "keybind"}, key)
end

function FlyoutListButton:ClearBindings() -- removes all keys bound to the given button
	local binding = "CLICK "..self:GetName()..":LeftButton"
	while GetBindingKey(binding) do
		SetBinding(GetBindingKey(binding), nil)
		self:SetKey(nil)
	end
end

function FlyoutListButton:FreeKey(key) -- unbinds the given key from all other buttons
	for _, arrowBtn in pairs(FbcArrowButtons) do
		for i, v in ipairs(arrowBtn.FlyoutListFrame.ButtonList) do
			if v.keybind == key then
				v:SetKey(nil)
			end
		end
	end
end

--tooltip

function FlyoutListButton:UpdateTooltipSpell()
	if (self.id) then --tooltipValue
		GameTooltip:SetSpellByID(self.id)
	end
end
function FlyoutListButton:UpdateTooltipItem()
	if (self.tooltipValue) then
		GameTooltip:SetItemByID(self.tooltipValue)
	end
end
function FlyoutListButton:UpdateTooltipCompanion()
	if (self.id) then
		GameTooltip:SetSpellByID(self.id)
	elseif (self.tooltipValue) then
		GameTooltip:SetHyperlink(self.tooltipValue)
	end
end
function FlyoutListButton:UpdateTooltipMacro()
	if (self.tooltipValue) then
		GameTooltip:SetText(self.tooltipValue, 1.0, 1.0, 1.0)
	end
end
function FlyoutListButton:UpdateTooltipEquipmentset()
	if (self.tooltipValue) then
		GameTooltip:SetEquipmentSet(self.tooltipValue)
	end
end
function FlyoutListButton:UpdateTooltipBattlepet()
	if (self.value) and (self.tooltipValue) then -- self.tooltipValue is hyperlinklink
		--speciesID, customName, level, xp, maxXp, displayID, isFavorite, name, icon, petType, creatureID, sourceText, description, isWild, canBattle, tradable, unique
		local speciesID, customName, level = C_PetJournal.GetPetInfoByPetID(self.value)
		local _, _, _, breedQuality, maxHealth, power, speed, _ = strsplit(":", self.tooltipValue)
		BattlePetToolTip_Show(speciesID, level, tonumber(breedQuality), maxHealth, power, speed, customName) 
	end
end
function FlyoutListButton:OnEnter()
	if (self.GetHotkey) then
		FbcLibKeyBound:Set(self)
	end
	
	if (GetCVar("UberTooltips") == "1") then
		GameTooltip_SetDefaultAnchor(GameTooltip, self)
	else
		local bar = self.actnBtn:GetParent()
		if (bar == MultiBarBottomRight or bar == MultiBarRight or bar == MultiBarLeft) then
			GameTooltip:SetOwner(self, "ANCHOR_LEFT")
		else
			GameTooltip:SetOwner(self, "ANCHOR_RIGHT")
		end
	end
	
	self:UpdateTooltip()

	-- anti fading stuff
	local parent = self.actnBtn
	while parent do
		local bar = parent:GetParent()
		if not(bar) or bar == UIParent then
			break
		elseif bar:GetAlpha() ~= 1 then
			ParentAlpha = {}
			ParentAlpha.bar = bar
			ParentAlpha.alpha = bar:GetAlpha()
			bar:SetAlpha(1)
			break
		end
		parent = bar
	end
end

function FlyoutListButton:OnLeave()
	GameTooltip:Hide()
	BattlePetTooltip:Hide()
	
	-- anti fading stuff
	if ParentAlpha and ParentAlpha.bar then
		ParentAlpha.bar:SetAlpha(ParentAlpha.alpha)
		ParentAlpha = nil
	end
end

function FlyoutListButton:SetTooltipSpell()
	self.tooltipValue = self.value
end
function FlyoutListButton:SetTooltipItem()
	self.tooltipValue = self.value --itemId
end
function FlyoutListButton:SetTooltipMacro()
	self.tooltipValue = self.value
end
function FlyoutListButton:SetTooltipEquipmentset()
	self.tooltipValue = self.value
end
function FlyoutListButton:SetTooltipCompanion()
	self.tooltipValue = nil
	if self.id then
		self.tooltipValue = "spell:"..self.id
	end
end
function FlyoutListButton:SetTooltipBattlepet()
	self.tooltipValue = C_PetJournal.GetBattlePetLink(self.value)
end

--texture

function FlyoutListButton:GetTextureSpell()
	local texture = nil
	local value, subValue, id = self.value, self.subValue, self.id

	if id then
		texture = GetSpellTexture(id)
	end
	if not(texture) then
		texture = GetSpellTexture(value) 
	end
	return texture
end
function FlyoutListButton:GetTextureItem()
	return GetItemIcon(self.value)
end
function FlyoutListButton:GetTextureMacro()
	local _, texture = GetMacroInfo(self.value)
	return texture
end
function FlyoutListButton:GetTextureEquipmentset()
	return 'Interface/Icons/'..(GetEquipmentSetInfoByName(self.value) or '')
end
function FlyoutListButton:GetTextureCompanion()
	local _, _, _, texture = FlyoutButton_GetCompanionInfoByName(self.value, self.subValue)
	return texture
end
function FlyoutListButton:GetTextureBattlepet()
	local _, _, _, _, _, _, _, _, icon = C_PetJournal.GetPetInfoByPetID(self.value)
	return icon
end
function FlyoutListButton:UpdateTexture()
	local icon = self.icon
	if not(self.command) then
		icon:Hide()
		self.cooldown:Hide()
		self:SetNormalTexture(QUICK_SLOT)
		return
	end

	local texture = self:GetTexture()

	if (texture) then
		icon:SetTexture(texture)
		icon:SetVertexColor(1.0, 1.0, 1.0, 1.0)
		icon:Show()
		if FbcShowBorders then
			self:SetNormalTexture(QUICK_SLOT_2)
		else
			self:SetNormalTexture("") --remove border
		end
	else
		icon:SetTexture(QUESTION_MARK)
		icon:SetVertexColor(1.0, 1.0, 1.0, 0.5)
		icon:Show()
		self:SetNormalTexture(QUICK_SLOT_2)		
	end
end

--checked

function FlyoutListButton:UpdateCheckedSpell()
	local result = false
	local id = self.id
	if id then
		result = IsCurrentSpell(id) or IsAutoRepeatSpell(id)
	end
	self:SetChecked(result)
end
function FlyoutListButton:UpdateCheckedItem()
	self:SetChecked(IsCurrentItem(self.value))
end
function FlyoutListButton:UpdateCheckedMacro()
	self:SetChecked(false)
end
function FlyoutListButton:UpdateCheckedEquipmentset()
	self:SetChecked(false) --todo get somehow is equipped
end
function FlyoutListButton:UpdateCheckedCompanion()
	local value, subValue = self.value, self.subValue
	local _, _, _, _, result = FlyoutButton_GetCompanionInfoByName(value, subValue)
	local spellName = UnitCastingInfo("player")
	self:SetChecked(result or spellName == value)
end
function FlyoutListButton:UpdateCheckedBattlepet()
	self:SetChecked(self.value == C_PetJournal.GetSummonedPetGUID())
	--self:SetChecked(false)
end

--equipped

function FlyoutListButton:UpdateEquippedItem()
	local border = self.border

	if IsEquippedItem(self.value) then
		border:SetVertexColor(0, 1.0, 0, 0.35)
		border:Show()
	else
		border:Hide()
	end
end
function FlyoutListButton:UpdateEquippedNotItem()
	self.border:Hide()
end

--cooldown

function FlyoutListButton:UpdateCooldownSpell()
	--print(tostring(self.command)..", "..tostring(self.value)..", "..tostring(self.subValue)..", "..tostring(self.id))
	local start, duration, enable = GetSpellCooldown(self.value) --GetSpellCooldown(id)
	if start then
		CooldownFrame_SetTimer(self.cooldown, start, duration, enable)
	else
		self:UpdateCooldownNone()
	end
end
function FlyoutListButton:UpdateCooldownItem()
	CooldownFrame_SetTimer(self.cooldown, GetItemCooldown(self.value))
end
function FlyoutListButton:UpdateCooldownMacro()
	local index = self.subValue --GetMacroIndexByName(value)
	if (index) then
		local name, rank = GetMacroSpell(index)
		if (name) then
			local spell = FlyoutButton_GetFullSpellName(name, rank)
			local start, duration, enable = GetSpellCooldown(spell)
			if start then
				CooldownFrame_SetTimer(self.cooldown, start, duration, enable)
			else
				self:UpdateCooldownNone()
			end
		end
	end
end
function FlyoutListButton:UpdateCooldownNone()
	CooldownFrame_SetTimer(self.cooldown, 0, 0, 0)
	self.cooldown:Hide()
end

--usable

function FlyoutListButton:GetUsableSpell()
	return IsUsableSpell(self.id) --self.value
end
function FlyoutListButton:GetUsableItem()
	return IsUsableItem(self.value)
end
function FlyoutListButton:GetUsableMacro()
	return true
end
function FlyoutListButton:GetUsableEquipmentset()
	return not(InCombatLockdown())
end
function FlyoutListButton:GetUsableComapnion()
	return self.subValue == "MOUNT" and IsOutdoors()
end
function FlyoutListButton:GetUsableBattlepet()
	return true
end
function FlyoutListButton:UpdateUsable()
	local isUsable, notEnoughMana = self:GetUsable()

	local icon = self.icon
	local ntexture = self.ntexture
	if (isUsable) then
		icon:SetVertexColor(1.0, 1.0, 1.0)
		ntexture:SetVertexColor(1.0, 1.0, 1.0)
	elseif (notEnoughMana and not(self.subValue == "MOUNT")) then
		icon:SetVertexColor(0.5, 0.5, 1.0)
		ntexture:SetVertexColor(0.5, 0.5, 1.0)
	else
		icon:SetVertexColor(0.4, 0.4, 0.4)
		ntexture:SetVertexColor(1.0, 1.0, 1.0)
	end	
end

--update text

function FlyoutListButton:UpdateTextMacro()
	-- name, iconTexture, body, isLocal = GetMacroInfo("name" or macroSlot)
	name = GetMacroInfo(self.subValue) --by slot
	self.name:SetText(name)
end
function FlyoutListButton:UpdateTextEquipmentset()
	self.name:SetText(self.value)
end

--range

function FlyoutListButton:InRangeSpell()
	return IsSpellInRange(self.value, "target")
end
function FlyoutListButton:InRangeItem()
	return IsItemInRange(self.value, "target")
end
function FlyoutListButton:InRangeMacro()
	return nil
end
function FlyoutListButton:InRangeEquipmentset()
	return nil
end
function FlyoutListButton:InRangeCompanion()
	return nil
end
function FlyoutListButton:InRangeBattlepet()
	return nil
end
function FlyoutListButton:UpdateRange()
	local hotkey = self.hotkey
	local valid = self:InRange()
	if ( hotkey:GetText() == RANGE_INDICATOR ) then
		if ( valid == 0 ) then
			hotkey:Show()
			hotkey:SetVertexColor(1.0, 0.1, 0.1)
		elseif ( valid == 1 ) then
			hotkey:Show()
			hotkey:SetVertexColor(0.6, 0.6, 0.6)
		else
			hotkey:Hide()
		end
	else
		if ( valid == 0 ) then
			hotkey:SetVertexColor(1.0, 0.1, 0.1)
		else
			hotkey:SetVertexColor(0.6, 0.6, 0.6)
		end
	end
end

-- count

local function GetCountText(count)
	if (count > 999) then
		return "*"
	else
		return count
	end
end
function FlyoutListButton:UpdateCountSpell()
	local text = ""
	if self.id then
		local charges, maxCharges, chargeStart, chargeDuration = GetSpellCharges(self.id)
		if (maxCharges) and (charges) and (maxCharges > 1) then
			text = GetCountText(charges)
		end
	end
	self.count:SetText(text)
	if (text) and (text ~= "") then
		self.count:Show()
	end
end
function FlyoutListButton:UpdateCountItem()
	local charges = GetItemCount(self.value, false, true)
	if charges > 1 then
		charges = GetCountText(charges)
	else
		charges = ""
	end
	self.count:SetText(charges)
	if (charges) and (charges ~= "") then
		self.count:Show()
	end
end

-- glow

function FlyoutListButton:SpellGlowShow(...)
	local arg1 = ...
	if self.id and self.id == arg1 then
		ActionButton_ShowOverlayGlow(self)
	end
end
function FlyoutListButton:MacroGlowShow(...)
	local arg1 = ...
	if self.subValue  then
		local _, _, spellId = GetMacroSpell(self.subValue) --index
		if spellId and spellId == arg1 then
			ActionButton_ShowOverlayGlow(self)
		end
	end
end
function FlyoutListButton:SpellGlowHide(...)
	local arg1 = ...
	if self.id and self.id == arg1 then
		ActionButton_HideOverlayGlow(self)
	end
end
function FlyoutListButton:MacroGlowHide(...)
	local arg1 = ...
	if self.subValue then
		local _, _, spellId = GetMacroSpell(self.subValue) --index
		if spellId and spellId == arg1 then
			ActionButton_HideOverlayGlow(self)
		end
	end
end

-- click

function FlyoutListButton:Click(button)
	if (self.command == "battlepet") and self.value then
		C_PetJournal.SummonPetByGUID(self.value)
	end
end

-- common

function FlyoutListButton:Set(command, value, subValue, id)
	self.command, self.value, self.subValue, self.id = command, value, subValue, id
	self:SetAttribute("type", "")
	self:SetAttribute("spell", "")
	self:SetAttribute("item", "")
	self:SetAttribute("macro", "")
	self:SetAttribute("clickbutton", "")
	self:SetAttribute("_onclick", "")

	if command then
		if (command == "spell") then
			self:SetAttribute("type", command)
			if (subValue == "MOUNT") then --everything on their own, mount spell by id not works
				for k, v in pairs(preset_by_companion) do
					self[k] = self[v]
				end

				self:SetAttribute(command, value)
			else
				for k, v in pairs(preset_by_spell) do
					self[k] = self[v]
				end

				self:SetAttribute(command, value) --id
			end
		elseif command == "item" then
			for k, v in pairs(preset_by_item) do
				self[k] = self[v]
			end

			local itemname = GetItemInfo(value)
			self:SetAttribute("type", command)
			self:SetAttribute("item", itemname)
		elseif (command == "macro") then
			for k, v in pairs(preset_by_macro) do
				self[k] = self[v]
			end

			self:SetAttribute("type", command)
			self:SetAttribute(command, value)
		elseif (command == "equipmentset") then
			for k, v in pairs(preset_by_equipmentset) do
				self[k] = self[v]
			end

			self:SetAttribute("type", "macro")
			self:SetAttribute("macrotext", "/equipset "..value)
		elseif (command == "battlepet") then
			for k, v in pairs(preset_by_battlepet) do
				self[k] = self[v]
			end

			self:SetAttribute("type", "click")
			self:SetAttribute("clickbutton", self)
			--local speciesID, customName, level, xp, maxXp, displayID, petName, petIcon, petType, creatureID = C_PetJournal.GetPetInfoByPetID(value)
		else
			self:ClearHandlers()
		end
	else
		self:ClearHandlers()
	end
	
	self:UpdateButton()
	self:Show()
	if command then
		self:RegisterEvents()
	else
		self:UnregisterEvents()
	end
end

function FlyoutListButton:PreClick(button)
	if InCombatLockdown() then
		ClearCursor()
		return
	end
	
	StoredCursor = {}
	local command, value, subValue, id = FlyoutButton_GetCursorValues()
	if button == "LeftButton" then
		if command and FbcSettingsMode then
			StoredCursor.command, StoredCursor.value, StoredCursor.subValue, StoredCursor.id = command, value, subValue, id
			self:SetAttribute("type", nil) --to avoid click event (spell cast)
		end
	elseif button == "RightButton" and FbcRightClickSet then
		StoredCursor.prevCommand = self:GetAttribute("type")
		self:SetAttribute("type", nil) --to avoid click event (spell cast)
	end
end

function FlyoutListButton:PostClick(button)
	if FbcLeftClickSet and self:GetAttribute("override_pending") 
	and (self.actnBtn:GetAttribute("fbc_tooltip_source") == self:GetName()) then
		FlyoutArrowButton_SetActionButtonTexture(self.actnBtn.icon)
	end
		
	if InCombatLockdown() then
		return
	end
	
	if button == "LeftButton" and StoredCursor.command then
		FlyoutButton_SetCursor(StoredCursor.command, StoredCursor.value, StoredCursor.subValue, StoredCursor.id)
		StoredCursor = {}
		self:OnReceiveDrag()
	end
	
	if (self:GetAttribute("override_pending") and (self.actnBtn:GetAttribute("fbc_tooltip_source") == self:GetName()))
	or (button == "RightButton" and FbcRightClickSet) then
		self:SetAttribute("override_pending", nil)
		local actnBtn = self.actnBtn
		actnBtn:SetAttribute("fbc_tooltip_source", nil)
		if StoredCursor.prevCommand then
			self:SetAttribute("type", StoredCursor.prevCommand)
		end
		local actBtnName = actnBtn:GetName()
		--print("Post click", actBtnName)
		if self.command then
			-- frame._state_action for BT4 ?
			if actBtnName:match("^ButtonForge%d+$") and BFButton and BFButton.OnReceiveDrag then
				FlyoutButton_SetCursor(self.command, self.value, self.subValue, self.id)
				BFButton.OnReceiveDrag(actnBtn)
				ClearCursor()
			elseif actBtnName:match("^BT4Button%d+$") or actBtnName:match("^ElvUI_Bar%dButton%d+$") then
				local atype, action = actnBtn:GetAction()
				if atype == "action" then
					FlyoutButton_SetCursor(self.command, self.value, self.subValue, self.id)
					PlaceAction(action)
					actnBtn:UpdateAction(true)
					ClearCursor()
				end
			elseif actBtnName:find("TinyExtraBars") and actnBtn.OnReceiveDrag then
				FlyoutButton_SetCursor(self.command, self.value, self.subValue, self.id)
				actnBtn:OnReceiveDrag()
				ClearCursor()
			elseif actnBtn.action and actnBtn.action <= 120 then
				FlyoutButton_SetCursor(self.command, self.value, self.subValue, self.id)
				PlaceAction(actnBtn.action)
				ActionButton_Update(actnBtn) 				
				ClearCursor()
			end
		end
		self:SetChecked(nil)
	else
		self:SetAttribute("override_pending", nil)
	end
end

function FlyoutListButton:SaveCommand(command, value, subValue, id)
	local actnBtnName = self.actnBtnName
	local currentSet = GetActiveSpecGroup()
	if command then
		FlyoutButtonES:Set({actnBtnName, currentSet, self.index, "command"}, command)
		FlyoutButtonES:Set({actnBtnName, currentSet, self.index, "value"}, value)
		FlyoutButtonES:Set({actnBtnName, currentSet, self.index, "subValue"}, subValue)
		FlyoutButtonES:Set({actnBtnName, currentSet, self.index, "id"}, id)
	else
		--empty tables to nil
		FlyoutButtonES:Set({actnBtnName, currentSet, self.index}, nil)
	end
end

function FlyoutListButton:OnReceiveDrag()
	if InCombatLockdown() then
		ClearCursor()
		return
	end

	local command, value, subValue, id = FlyoutButton_GetCursorValues()

	ClearCursor()
	if self.command then
		FlyoutButton_SetCursor(self.command, self.value, self.subValue, self.id)
	end

	self:Set(command, value, subValue, id)
	self:SaveCommand(command, value, subValue, id)
	
	self:GetParent():GetParent():UpdateArrow()
end

function FlyoutListButton:OnDragStart()
	if not(FbcSettingsMode) then
		return
	end
	
	local arrowBtn = self:GetParent():GetParent()

	if self.command then
		FlyoutButton_SetCursor(self.command, self.value, self.subValue, self.id)
	end
	
	FlyoutButtonES:Set({self.actnBtnName, GetActiveSpecGroup(), self.index}, {})
	
	self:Set(nil, nil, nil, nil)
	self:SaveCommand(nil, nil, nil, nil)
	
	arrowBtn:UpdateArrow()
end

function FlyoutListButton:HideButton()
	self.tooltipValue = nil
	self:SetAttribute("type", "none")
	self.icon:Hide()
	self.cooldown:Hide()
	self.border:Hide()
	self.count:Hide()
	self.name:SetText("")
	self.hotkey:Hide()
	self:SetNormalTexture(QUICK_SLOT)
	self:SetChecked(0)
	self:Hide()
end

function FlyoutListButton:UpdateButton()
	self:SetTooltip()
	self:UpdateTexture()
	self:UpdateCooldown()
	self:UpdateChecked()
	self:UpdateEquipped()
	self:UpdateUsable()
	self:UpdateText()
	self:UpdateCount()
	if self.command then
		self.rangeTimer = -1
	else
		self.rangeTimer = nil
	end
end

local function EventCooldownUpdate(self, ...)
	self:UpdateCooldown()
end

local function EventCheckedUpdate(self, ...)
	self:UpdateChecked()
end

local function EventEquippedUpdate(self, ...)
	self:UpdateEquipped()
	self:UpdateText()
	self:UpdateCooldown()
	self:UpdateChecked()
end

local function EventUsableUpdate(self, ...)
	self:UpdateUsable()
	self:UpdateText()
	self:UpdateCooldown()
end

local function EventMacroUpdate(self, ...)
	if InCombatLockdown() then
		return
	end

	if self.command == "macro" then
		--command = "macro", value = macro name, subValue = macro index
		local command, subValue = self.command, self.subValue
		local value = GetMacroInfo(subValue)
		self:Set(command, value, subValue, nil)
		self:SaveCommand(command, value, subValue, nil)
		self.arrBtn:UpdateArrow()
	end
end

local function EventRangeUpdate(self, ...)
	self:UpdateRange()
	self.rangeTimer = TOOLTIP_UPDATE_TIME
end

local function EventChargesUpdate(self, ...)
	self:UpdateCount()
end

local function EventSpellsChanged(self, ...)
	self:UpdateButton()
end

local function EventGlowShow(self, ...)
	self:GlowShow(...)
end

local function EventGlowHide(self, ...)
	self:GlowHide(...)
end

local function EventLeavingCombat(self, ...)
	if self:GetAttribute("override_pending") then
		self:PostClick(nil) --"RightButton"
	end
end

function FlyoutListButton:OnUpdate(elapsed)
	local rangeTimer = self.rangeTimer
	if ( rangeTimer ) then
		rangeTimer = rangeTimer - elapsed

		if ( rangeTimer <= 0 ) then
			self:UpdateRange()
			rangeTimer = TOOLTIP_UPDATE_TIME
		end
		
		self.rangeTimer = rangeTimer
	end
end

function FlyoutListButton:OnEvent(event, ...)
	--print("flb "..tostring(event))
	if self.EventHandlersTable[event] then
		self.EventHandlersTable[event](self, ...)
	end
end

function FlyoutListButton:RegisterEvents()
	--event table
	self.EventHandlersTable = {
		--cooldown
		["SPELL_UPDATE_COOLDOWN"] 			= EventCooldownUpdate,
		["BAG_UPDATE_COOLDOWN"] 			= EventCooldownUpdate,
		["ACTIONBAR_UPDATE_COOLDOWN"] 		= EventCooldownUpdate,
		["UPDATE_SHAPESHIFT_COOLDOWN"] 		= EventCooldownUpdate,
		--checked
		["TRADE_SKILL_SHOW"] 				= EventCheckedUpdate,
		["TRADE_SKILL_CLOSE"] 				= EventCheckedUpdate,
		["ARCHAEOLOGY_TOGGLE"] 				= EventCheckedUpdate,
		["ARCHAEOLOGY_CLOSED"] 				= EventCheckedUpdate,
		["COMPANION_UPDATE"] 				= EventCheckedUpdate,
		["CURRENT_SPELL_CAST_CHANGED"] 		= EventCheckedUpdate,
		["ACTIONBAR_UPDATE_STATE"] 			= EventCheckedUpdate,
		["PLAYER_ENTER_COMBAT"] 			= EventCheckedUpdate,
		["PLAYER_LEAVE_COMBAT"] 			= EventCheckedUpdate,
		["START_AUTOREPEAT_SPELL"] 			= EventCheckedUpdate,
		["STOP_AUTOREPEAT_SPELL"] 			= EventCheckedUpdate,
		["UPDATE_BONUS_ACTIONBAR"] 			= EventCheckedUpdate,
		["ACTIONBAR_PAGE_CHANGED"] 			= EventCheckedUpdate,
		--equipment
		["PLAYER_EQUIPMENT_CHANGED"] 		= EventEquippedUpdate,
		--usable
		["BAG_UPDATE"] 						= EventUsableUpdate,
		--["UNIT_INVENTORY_CHANGED"] 		= EventUsableUpdate,
		["SPELL_UPDATE_USABLE"] 			= EventUsableUpdate,
		["PLAYER_CONTROL_LOST"] 			= EventUsableUpdate,
		["PLAYER_CONTROL_GAINED"] 			= EventUsableUpdate,
		["UPDATE_BONUS_ACTIONBAR"] 			= EventUsableUpdate,
		["ACTIONBAR_UPDATE_USABLE"] 		= EventUsableUpdate,
		["VEHICLE_UPDATE"] 					= EventUsableUpdate,
		["UPDATE_WORLD_STATES"] 			= EventUsableUpdate,
		--macro
		["UPDATE_MACROS"]					= EventMacroUpdate,
		--range
		["PLAYER_TARGET_CHANGED"]			= EventRangeUpdate,
		--charges
		["SPELL_UPDATE_CHARGES"]			= EventChargesUpdate,
		--spells
		["ACTIVE_TALENT_GROUP_CHANGED"]		= EventSpellsChanged,
		["SPELLS_CHANGED"]					= EventSpellsChanged,
		["LEARNED_SPELL_IN_TAB"] 			= EventSpellsChanged,
		["PLAYER_GUILD_UPDATE"] 			= EventSpellsChanged,
		["PLAYER_SPECIALIZATION_CHANGED"] 	= EventSpellsChanged,
		["COMPANION_LEARNED"] 				= EventSpellsChanged,
		["UPDATE_SHAPESHIFT_FORM"] 			= EventSpellsChanged,
		--glow
		["SPELL_ACTIVATION_OVERLAY_GLOW_SHOW"] = EventGlowShow,
		["SPELL_ACTIVATION_OVERLAY_GLOW_HIDE"] = EventGlowHide,
		--regen to save overrides
		["PLAYER_REGEN_ENABLED"] 			= EventLeavingCombat,
	}
	
	for k, _ in pairs(self.EventHandlersTable) do
		self:RegisterEvent(k)
	end

	self:SetScript("OnEvent", FlyoutListButton.OnEvent)
	self:SetScript("OnUpdate", FlyoutListButton.OnUpdate)
end

function FlyoutListButton:UnregisterEvents()
	if self.EventHandlersTable then
		for k, _ in pairs(self.EventHandlersTable) do
			self:UnregisterEvent(k)
		end
	end
	self.EventHandlersTable = {}
	self.rangeTimer = nil
	self.hotkey:Hide()
end

